#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <descrip.h>
#include <starlet.h>
#include <rms.h>
#include <ssdef.h>
#include <iodef.h>
#include <iosadef.h>
#include <fibdef.h>
#include <atrdef.h>
#include <gen64def.h>

#pragma pointer_size save
#pragma pointer_size 64
typedef void *void_ptr64;
#pragma pointer_size restore

#include "high_res_timer.h"

struct FAT
{
    unsigned long fat$v_rtype: 4;
    unsigned long fat$v_fileorg: 4;
    unsigned char fat$b_rattrib;
    unsigned short fat$w_rsize;
    unsigned short fat$w_hiblkh,fat$w_hiblkl;
    unsigned short fat$w_efblkh,fat$w_efblkl;
    unsigned short fat$w_ffbyte;
    unsigned char fat$b_bktsize;
    unsigned char fat$b_vfcsize;
    unsigned short fat$w_maxrec;
    unsigned short fat$w_defext;
    unsigned short fat$w_gbc;
    unsigned char fill_1[8];
    unsigned short fat$w_versions;
};

static long long offset(int blkno, int bytno)
{
    return (blkno - 1) * 512L + bytno;
}

static __align(13) char realbuf[819200];
static __align(13) union { iosa fastiosb; char dummy[8192]; } fast;

static void test(const char *fnm, int expect, int blksiz)
{
    long stat;
    unsigned short iosb[4];
    struct FAB fab;
    struct NAM nam;
    struct dsc$descriptor_s devdesc;
    short chan;
    struct fibdef fib;
    long fibdesc[] = { FIB$K_LENGTH, (long)&fib };
    struct FAT fat;
    struct atrdef atr[] = { {ATR$S_RECATTR, ATR$C_RECATTR, &fat},{ 0, 0, NULL } };
    struct _generic_64 realbufh, fasth;
    void_ptr64 realbufptr, *fastptr;
    long long realbuflen, fastlen;
    long long fh;
    long eof;
    short ffb;
    int adjblksiz;
    char expfnm[NAM$C_MAXRSS];
    int ix, len, n, i;
    TIMECOUNT_T t1, t2;
    t1 = GET_TIMECOUNT;
    fab = cc$rms_fab;
    nam = cc$rms_nam;
    fab.fab$l_fna = (char *)fnm;
    fab.fab$b_fns = strlen(fnm);
    fab.fab$b_fac = FAB$M_GET;
    fab.fab$l_nam = &nam;
    nam.nam$b_rss = sizeof(expfnm) - 1;
    nam.nam$l_rsa = expfnm;
    stat = sys$open(&fab, 0, 0);
    if(!(stat & 1))
    {
        printf("sys$open stat = %d\n", stat);
    }
    stat = sys$close(&fab, 0, 0);
    if(!(stat & 1))
    {
       printf("sys$close stat = %d\n", stat);
    }
    devdesc.dsc$b_dtype = DSC$K_DTYPE_T;
    devdesc.dsc$b_class = DSC$K_CLASS_S;
    devdesc.dsc$w_length = strlen(expfnm);
    devdesc.dsc$a_pointer = expfnm;
    stat = sys$assign(&devdesc, &chan, 0, 0, 0);
    if(!(stat & 1))
    {
       printf("sys$assign stat = %d\n", stat);
    }
    memset(&fib, 0, sizeof(fib));
    fib.fib$w_fid[0] = nam.nam$w_fid[0];
    fib.fib$w_fid[1] = nam.nam$w_fid[1];
    fib.fib$w_fid[2] = nam.nam$w_fid[2];
    stat = sys$qiow(0, chan, IO$_ACCESS | IO$M_ACCESS, iosb, 0, 0, &fibdesc, 0, 0, 0, &atr, 0);
    if(!(stat & 1))
    {
       printf("sys$qiow io$_access stat = %d\n", stat);
    }
    eof = (fat.fat$w_efblkh << 16) | fat.fat$w_efblkl;
    ffb = fat.fat$w_ffbyte;
    stat = sys$create_bufobj_64(realbuf, sizeof(realbuf), 0, 0, &realbufptr, &realbuflen, &realbufh);
    if(!(stat & 1))
    {
       printf("sys$create_bufobj_64 realbuf stat = %d\n", stat);
    }
    stat = sys$create_bufobj_64(&fast, sizeof(fast), 0, 0, &fastptr, &fastlen, &fasth);
    if(!(stat & 1))
    {
       printf("sys$create_bufobj_64 fast stat = %d\n", stat);
    }
    stat = sys$io_setup (IO$_READVBLK, &realbufh, &fasth, 0, 0, &fh);
    if(!(stat & 1))
    {
       printf("sys$io_setup stat = %d\n", stat);
    }
    ix = 1;
    n = 0;
    for(;;)
    {
        if(ix > eof) break;
        if(offset(ix, blksiz) <= offset(eof, ffb))
        {
            adjblksiz = blksiz;
        }
        else
        {
            adjblksiz = offset(eof, ffb) - offset(ix, 0);
        }
        stat = sys$io_performw(fh, chan, &fast.fastiosb, realbuf, adjblksiz, ix);
        if(fast.fastiosb.iosa$l_status == SS$_ENDOFFILE)
        {
            break;
        }
        else if(!(stat & 1))
        {
           printf("sys$io_performw stat = %d\n", stat);
        }
        else if(!(fast.fastiosb.iosa$l_status & 1))
        {
           printf("sys$io_performw iosb_status = %d\n", fast.fastiosb.iosa$l_status);
        }
        len = fast.fastiosb.iosa$ih_count;
        for(i = 0; i < len; i++)
        {
            if(realbuf[i] == '\n') n++;
        }
        ix = ix + blksiz / 512;
    }
    if(n != expect)
    {
        printf("%d lines\n", n);
    }
    stat = sys$io_cleanup(fh);
    if(!(stat & 1))
    {
       printf("sys$io_cleanup stat = %d\n", stat);
    }
    stat = sys$delete_bufobj(&realbufh);
    if(!(stat & 1))
    {
       printf("sys$delete_bufobj realbuf stat = %d\n", stat);
    }
    stat = sys$delete_bufobj(&fasth);
    if(!(stat & 1))
    {
       printf("sys$delete_bufobj fast stat = %d\n", stat);
    }
    stat = sys$qiow(0, chan, IO$_DEACCESS, iosb, 0, 0, &fibdesc, 0, 0, 0, &atr, 0);
    if(!(stat & 1))
    {
       printf("sys$qiow io$_deaccess stat = %d\n", stat);
    }
    stat = sys$dassgn(chan);
    if(!(stat & 1))
    {
       printf("sys$dassgn stat = %d\n", stat);
    }
    t2 = GET_TIMECOUNT;
    printf("sys$io_performw (block size %d) : %d ms\n", blksiz, (int)((t2 - t1) * 1000 / UNITS_PER_SECOND));
}

int main(int argc, char *argv[])
{
    int blksiz[] = { 512, 5120, 51200, 512000 };
    int i, j;
    for(j = 0; j < 4; j++)
    {
        for(i = 0; i < 10; i++)
        {
            test(argv[1], atoi(argv[2]), blksiz[j]);
        }
    }
    return 0;
}
